
NAME = 'peony'

# 系统返回码
# admin用户验证失败
RET_ADMIN_AUTH_FAIL = -20000


# 管理员命令
# 获取运行状态统计
CMD_ADMIN_SERVER_STAT = 20000

# 停止整个server
CMD_ADMIN_STOP = 21003


DEFAULT_CONFIG = dict(
    HOST='127.0.0.1',
    PORT=9250,

    # 监听TCP
    TCP=True,

    # 监听UDP
    UDP=False,

    # 启动的group列表。group_id必须为数字
    GROUP_LIST=[dict(id=1, count=1)],

    # 通过task路由group_id
    GROUP_ROUTER=lambda task: 1,

    DEBUG=False,

    # box class
    BOX_CLASS='netkit.box.Box',

    # task class
    TASK_CLASS='maple.Task',

    # master class
    MASTER_CLASS='peony.master.Master',

    # proxy class
    PROXY_CLASS='peony.proxy.Proxy',

    # worker class
    WORKER_CLASS='peony.worker.Worker',

    # request class
    REQUEST_CLASS='peony.worker.Request',

    NAME=NAME,

    # 每个worker的消息队列的最大长度, <=0 代表无限
    TASK_QUEUE_MAX_SIZE=-1,

    # 一次性认领的任务数量最大值，<=0 代表无限
    TASK_CLAIMED_MAX_SIZE=-1,

    # backlog
    BACKLOG=256,

    # 客户端连接超时
    CLIENT_TIMEOUT=None,

    STOP_TIMEOUT=None,

    # 处理task超时(秒). 超过后会打印fatal日志. None 代表永不超时
    WORK_TIMEOUT=None,

    # 管理员，可以连接proxy获取数据
    # 管理员访问地址: 'admin.sock' or ('127.0.0.1', 9910)
    ADMIN_ADDRESS='admin.sock',
    ADMIN_USERNAME=None,
    ADMIN_PASSWORD=None,
)




"""
封装通过group_id来访问数据的功能
"""

import multiprocessing
import queue
from ..share.log import logger
from ..share.utils import safe_call


class TaskQueueManager(object):
    """
    通过group_id来区分的queue
    """

    max_size = None
    queue_dict = None

    def __init__(self, group_id_list, max_size=-1):
        self.max_size = max_size
        self.queue_dict = dict()
        for group_id in group_id_list:
            self.queue_dict[group_id] = multiprocessing.Queue(self.max_size)

    def get_queue(self, group_id):
        """
        获取对应的队列
        :param group_id:
        :return:
        """
        return self.queue_dict.get(group_id)

    def put(self, group_id, item):
        """
        加入item
        如果成功返回True，如果失败返回False
        :param group_id:
        :param item:
        :return:
        """
        try:
            self.queue_dict[group_id].put_nowait(item)
            return True
        except queue.Full:
            logger.error(
                'put item fail, queue is full. group_id: %s, max_size: %s',
                group_id, self.max_size
            )
            return False
        except:
            logger.error('put item fail, exc occur. group_id: %s', group_id, exc_info=True)
            return False

    def get(self, group_id):
        if self.empty(group_id):
            return None
        else:
            return self.queue_dict[group_id].get_nowait()

    def clear(self, group_id):
        """
        清空
        :param group_id:
        :return:
        """
        self.queue_dict.pop(group_id, None)

    def clear_all(self):
        """
        清空所有
        :return:
        """
        self.queue_dict.clear()

    def empty(self, group_id):
        """
        判断是否是空的
        :param group_id:
        :return:
        """
        return self.queue_dict[group_id].empty()

    def qsize(self, group_id):
        """
        某个队列的size
        :param group_id:
        :return:
        """
        return self.queue_dict[group_id].qsize()

    def close(self):
        """
        如果队列里面有数据的话结束不了进程
        :return:
        """

        for q in self.queue_dict.values():
            safe_call(q.close)



import os
import queue
import signal
import time
from pyglet.clock import Clock
import _thread
import setproctitle
from maple import constants as maple_constants

from ..share import constants
from ..share.log import logger
from ..share.utils import safe_call


class Worker:
    app = None
    group_id = None

    enabled = True

    # 业务可以通过clock来设置schedule
    clock = None

    _got_first_request = False

    # pipe 比 queue性能高一点，因为queue是在pipe的基础上加了锁
    # 但是最终还是得选择queue，因为pipe.recv是阻塞的，也就是每次poll完了只能读取一次
    # 这会导致clock的计算次数增加，没必要
    _task_queue = None

    # 工作进展
    _work_progress = None

    def __init__(self, app, group_id, task_queue):
        self.app = app
        self.group_id = group_id
        self._task_queue = task_queue

        self.clock = self._create_clock()

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(
            'worker:%s' % self.group_id
        ))

        self._handle_signals()

        self._on_start()

        # daemon==True，主线程不需要等待网络线程结束
        _thread.start_new_thread(self._monitor_work_timeout, ())

        while self.enabled:
            try:
                self._handle()
            except:
                logger.error('exc occur. worker: %s', self, exc_info=True)

        self._on_stop()

    def _create_clock(self):
        """
        创建clock
        """
        return Clock()

    def _handle(self):
        """
        主循环，等待网络消息，并进行处理
        """
        # 允许idle
        sleep_time = self.clock.get_sleep_time(True)
        # sleep_time is None 说明没有需要处理的，直接永久阻塞等待即可
        # 否则就要使用指定时间阻塞等待
        task_list = []

        # 第一个get要block
        block = True
        # 已经领取的task的数量，单独一个变量是为了性能考虑
        task_claimed_count = 0

        while self.enabled and not (
                task_claimed_count >= self.app.task_claimed_max_size > 0
        ):
            try:
                task_list.append(
                    self._task_queue.get(block=block, timeout=sleep_time)
                )

                task_claimed_count += 1
            except queue.Empty:
                # 说明已经没有了
                break
            except OSError:
                # 当term信号处理函数调用了queue.close()
                # 会在这里抛出异常 handle is closed
                # 直接break就好
                break
            except:
                # 说明出问题了
                logger.error('exc occur. worker: %s', self, exc_info=True)
                break
            finally:
                block = False

        self.app.events.activate_worker(self)
        for bp in self.app.blueprints:
            bp.events.app_activate_worker(self)

        for task in task_list:
            request = self.app.request_class(self, task)
            # 设置task开始处理的时间和信息
            self._work_progress = dict(
                begin_time=time.perf_counter(),
                request=request,
            )

            self._handle_request(request)

            self._work_progress = None

        safe_call(self.clock.tick, True)

        for bp in self.app.blueprints:
            bp.events.app_deactivate_worker(self)
        self.app.events.deactivate_worker(self)

    def _handle_request(self, request):
        """
        出现任何异常的时候，服务器不再主动关闭连接
        """
        if not request.is_valid:
            return False

        if request.task.cmd not in (maple_constants.CMD_CLIENT_CREATED, maple_constants.CMD_CLIENT_CLOSED) \
                and not request.view_func:
            logger.info('cmd invalid. request: %s' % request)
            return False

        if not self._got_first_request:
            self._got_first_request = True
            self.app.events.before_first_request(request)
            for bp in self.app.blueprints:
                bp.events.before_app_first_request(request)

        self.app.events.before_request(request)
        for bp in self.app.blueprints:
            bp.events.before_app_request(request)
        if request.blueprint:
            request.blueprint.events.before_request(request)

        if request.interrupted:
            # 业务要求中断
            return True

        view_func_exc = None

        if request.task.cmd == maple_constants.CMD_CLIENT_CREATED:
            self.app.events.create_client(request)
            for bp in self.app.blueprints:
                bp.events.create_app_client(request)
        elif request.task.cmd == maple_constants.CMD_CLIENT_CLOSED:
            self.app.events.close_client(request)
            for bp in self.app.blueprints:
                bp.events.close_app_client(request)
        else:
            try:
                request.view_func(request)
            except Exception as e:
                logger.error('view_func raise exception. e: %s, request: %s', e, request, exc_info=True)
                view_func_exc = e

        if request.blueprint:
            request.blueprint.events.after_request(request, view_func_exc)
        for bp in self.app.blueprints:
            bp.events.after_app_request(request, view_func_exc)
        self.app.events.after_request(request, view_func_exc)

        return True

    def _monitor_work_timeout(self):
        """
        监控task的耗时
        :return:
        """

        while True:
            time.sleep(1)

            work_progress = self._work_progress
            if work_progress:
                past_time = time.perf_counter() - work_progress['begin_time']
                if self.app.work_timeout is not None and past_time > self.app.work_timeout:
                    # 说明worker的处理时间已经太长了
                    logger.fatal('work timeout: %s / %s, request: %s',
                                 past_time, self.app.work_timeout, work_progress['request'])
                    # 强制从子线程退出整个进程
                    os._exit(-1)

    def _on_start(self):
        """
        当worker启动后
        可继承实现
        :return:
        """
        self.app.events.start_worker(self)
        for bp in self.app.blueprints:
            bp.events.start_app_worker(self)

    def _on_stop(self):
        """
        当worker停止前
        可继承实现
        :return:
        """
        for bp in self.app.blueprints:
            bp.events.stop_app_worker(self)
        self.app.events.stop_worker(self)

    def _handle_signals(self):
        def safe_stop_handler(signum, frame):
            self.enabled = False
            # 如果队列处于get中的话，会直接打断并抛出异常 OSError: handle is closed
            # 经测试不会影响其他worker对于该队列的读取
            safe_call(self._task_queue.close)

        # 安全停止
        signal.signal(signal.SIGTERM, safe_stop_handler)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

    def __repr__(self):
        return '<%s name: %s, group_id: %r>' % (
            type(self).__name__, self.app.name, self.group_id
        )





from twisted.internet.protocol import DatagramProtocol

from ...share.utils import safe_call
from ...share.log import logger
from ...share import constants


class ClientConnectionFactory(DatagramProtocol):

    def __init__(self, proxy):
        self.proxy = proxy

    def datagramReceived(self, data, address):
        # address: type: tuple, format: ('127.0.0.1', 62043)

        while data:
            task = self.proxy.app.task_class()
            ret = task.unpack(data)

            if ret == 0:
                # 说明要继续收
                logger.error('buffer incomplete. proxy: %s, ret: %d, read_buffer: %r',
                             self.proxy, ret, data)
                return
            elif ret > 0:
                # 收好了
                # 不能使用双下划线，会导致别的地方取的时候变为 _Gateway__raw_data，很奇怪
                task._raw_data = data[:ret]
                data = data[ret:]
                safe_call(self._on_read_complete, task, address)
                continue
            else:
                # 数据已经混乱了，全部丢弃
                logger.error('buffer invalid. proxy: %s, ret: %d, read_buffer: %r',
                             self.proxy, ret, data)
                data = b''
                return

    def _on_read_complete(self, task, address):
        self.proxy.stat_counter.client_req += 1

        group_id = self.proxy.app.group_router(task)
        self.proxy.stat_counter.add_worker_req(group_id)
        # 直接发送到对应的group_id
        if not self.proxy.app.master.push_task(group_id, task):
            self.proxy.stat_counter.add_discard_task(group_id)




from twisted.internet.protocol import Protocol, Factory, connectionDone
from twisted.internet import reactor

from ...share.utils import safe_call
from ...share.log import logger


class ClientConnectionFactory(Factory):

    def __init__(self, proxy):
        self.proxy = proxy

    def buildProtocol(self, addr):
        return ClientConnection(self, addr)


class ClientConnection(Protocol):
    _read_buffer = None

    # 过期timer
    _expire_timer = None

    def __init__(self, factory, address):
        # address: IPv4Address
        self.factory = factory
        self.address = address
        self._read_buffer = b''

    def connectionMade(self):
        self.transport.setTcpNoDelay(True)

        self.factory.proxy.stat_counter.clients += 1

        self._set_expire_callback()

    def connectionLost(self, reason=connectionDone):
        self._clear_expire_callback()

        self.factory.proxy.stat_counter.clients -= 1

    def dataReceived(self, data):
        """
        当数据接受到时
        :param data:
        :return:
        """
        self._read_buffer += data

        while self._read_buffer:
            # 因为task后面还是要用的
            task = self.factory.proxy.app.task_class()
            ret = task.unpack(self._read_buffer)
            if ret == 0:
                # 说明要继续收
                return
            elif ret > 0:
                # 收好了
                # 不能使用双下划线，会导致别的地方取的时候变为 _Gateway__raw_data，很奇怪
                task._raw_data = self._read_buffer[:ret]
                self._read_buffer = self._read_buffer[ret:]
                safe_call(self._on_read_complete, task)
                continue
            else:
                # 数据已经混乱了，全部丢弃
                logger.error('buffer invalid. proxy: %s, ret: %d, read_buffer: %r',
                             self.factory.proxy, ret, self._read_buffer)
                self._read_buffer = b''
                return

    def _on_read_complete(self, task):
        """
        完整数据接收完成
        :param task: 解析之后的task
        :return:
        """
        self.factory.proxy.stat_counter.client_req += 1
        self._set_expire_callback()

        group_id = self.factory.proxy.app.group_router(task)
        self.factory.proxy.stat_counter.add_worker_req(group_id)
        # 直接发送到对应的group_id
        if not self.factory.proxy.app.master.push_task(group_id, task):
            self.factory.proxy.stat_counter.add_discard_task(group_id)

    def _set_expire_callback(self):
        """
        注册超时之后的回调
        :return:
        """

        if self.factory.proxy.app.client_timeout is None:
            return

        self._clear_expire_callback()

        self._expire_timer = reactor.callLater(
            self.factory.proxy.app.client_timeout, self._expire_callback
        )

    def _clear_expire_callback(self):
        """
        清空超时之后的回调
        :return:
        """
        if self._expire_timer:
            self._expire_timer.cancel()
            self._expire_timer = None

    def _expire_callback(self):
        """
        能关闭的话，就关闭掉
        :return:
        """
        self._expire_timer = None

        if self.transport and self.transport.connected:
            self.transport.loseConnection()
